# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Union, Optional
from typing_extensions import Literal, TypeAlias

from .._models import BaseModel
from .transaction import Transaction

__all__ = [
    "TransactionListHistoryResponse",
    "Data",
    "DataDiff",
    "DataDiffAddAllocationDiffEntry",
    "DataDiffAddAllocationDiffEntryItem",
    "DataDiffAddAllocationDiffEntryItemUser",
    "DataDiffDeleteAllocationDiffEntry",
    "DataDiffDeleteAllocationDiffEntryItem",
    "DataDiffDeleteAllocationDiffEntryItemUser",
]


class DataDiffAddAllocationDiffEntryItemUser(BaseModel):
    id: str
    """FRAGMENT generated ID of the user"""

    external_id: Optional[str] = None
    """External ID of the user"""


class DataDiffAddAllocationDiffEntryItem(BaseModel):
    """Transaction allocation against an invoice."""

    amount: str
    """Amount to allocate in smallest currency unit as stringified bigint."""

    invoice_id: str
    """The invoice to allocate against."""

    type: Literal["invoice_payin", "invoice_payout"]
    """The type of allocation."""

    user: DataDiffAddAllocationDiffEntryItemUser


class DataDiffAddAllocationDiffEntry(BaseModel):
    item: DataDiffAddAllocationDiffEntryItem
    """Transaction allocation against an invoice."""

    op: Literal["add"]
    """An allocation was added"""


class DataDiffDeleteAllocationDiffEntryItemUser(BaseModel):
    id: str
    """FRAGMENT generated ID of the user"""

    external_id: Optional[str] = None
    """External ID of the user"""


class DataDiffDeleteAllocationDiffEntryItem(BaseModel):
    """Transaction allocation against an invoice."""

    amount: str
    """Amount to allocate in smallest currency unit as stringified bigint."""

    invoice_id: str
    """The invoice to allocate against."""

    type: Literal["invoice_payin", "invoice_payout"]
    """The type of allocation."""

    user: DataDiffDeleteAllocationDiffEntryItemUser


class DataDiffDeleteAllocationDiffEntry(BaseModel):
    item: DataDiffDeleteAllocationDiffEntryItem
    """Transaction allocation against an invoice."""

    op: Literal["delete"]
    """An allocation was deleted"""


DataDiff: TypeAlias = Union[DataDiffAddAllocationDiffEntry, DataDiffDeleteAllocationDiffEntry]


class Data(Transaction):
    """A versioned snapshot of a transaction"""

    diff: Optional[List[DataDiff]] = None
    """Allocation changes applied in this version.

    Absent on version 1 (initial creation). Each entry describes an allocation that
    was added or deleted.
    """

    version: Optional[int] = None  # type: ignore
    """Version number of this transaction snapshot."""


class TransactionListHistoryResponse(BaseModel):
    """Version history of a transaction"""

    data: List[Data]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Iterable
from typing_extensions import Literal, Required, TypeAlias, TypedDict

__all__ = [
    "InvoiceUpdateParams",
    "LineItemsUpdate",
    "LineItemsUpdateAddLineItemOperation",
    "LineItemsUpdateAddLineItemOperationUser",
    "LineItemsUpdateAddLineItemOperationUserID",
    "LineItemsUpdateAddLineItemOperationUserExternalID",
    "LineItemsUpdateAddLineItemOperationPrice",
    "LineItemsUpdateAddLineItemOperationTag",
    "LineItemsUpdateUpdateLineItemOperation",
    "LineItemsUpdateUpdateLineItemOperationPrice",
    "LineItemsUpdateDeleteLineItemOperation",
]


class InvoiceUpdateParams(TypedDict, total=False):
    line_items_update: Required[Iterable[LineItemsUpdate]]
    """List of line item operations to apply to the invoice"""

    version: Required[float]
    """The version of the invoice being updated.

    Must match the current version for the update to succeed.
    """


class LineItemsUpdateAddLineItemOperationUserID(TypedDict, total=False):
    id: Required[str]
    """FRAGMENT generated ID of the user"""


class LineItemsUpdateAddLineItemOperationUserExternalID(TypedDict, total=False):
    external_id: Required[str]
    """External ID of the user"""


LineItemsUpdateAddLineItemOperationUser: TypeAlias = Union[
    LineItemsUpdateAddLineItemOperationUserID, LineItemsUpdateAddLineItemOperationUserExternalID
]


class LineItemsUpdateAddLineItemOperationPrice(TypedDict, total=False):
    """Price breakdown. Provide amount, or unit_price + quantity, or all three."""

    amount: str
    """Total amount in smallest currency unit.

    Required if unit_price and quantity are not provided.
    """

    quantity: int
    """Number of units for this line item."""

    unit_price: str
    """Price per unit in smallest currency unit."""


class LineItemsUpdateAddLineItemOperationTag(TypedDict, total=False):
    """A key-value tag pair for metadata"""

    key: Required[str]
    """Tag key.

    Must be a valid safe string (no special characters like #, /, :). Max 50
    characters.
    """

    value: Required[str]
    """Tag value.

    Must be a valid safe string (no special characters like #, /, :). Max 200
    characters.
    """


class LineItemsUpdateAddLineItemOperation(TypedDict, total=False):
    """Operation to add a new line item to an invoice"""

    currency_code: Required[
        Literal[
            "ADA",
            "BTC",
            "DAI",
            "ETH",
            "SOL",
            "USDC",
            "USDT",
            "USDG",
            "EURC",
            "CADC",
            "CADT",
            "XLM",
            "UNI",
            "BCH",
            "LTC",
            "AAVE",
            "LINK",
            "MATIC",
            "PTS",
            "AED",
            "AFN",
            "ALL",
            "AMD",
            "ANG",
            "AOA",
            "ARS",
            "AUD",
            "AWG",
            "AZN",
            "BAM",
            "BBD",
            "BDT",
            "BGN",
            "BHD",
            "BIF",
            "BMD",
            "BND",
            "BOB",
            "BRL",
            "BSD",
            "BTN",
            "BWP",
            "BYR",
            "BZD",
            "CAD",
            "CDF",
            "CHF",
            "CLP",
            "CNY",
            "COP",
            "CRC",
            "CUC",
            "CUP",
            "CVE",
            "CZK",
            "DJF",
            "DKK",
            "DOP",
            "DZD",
            "EGP",
            "ERN",
            "ETB",
            "EUR",
            "FJD",
            "FKP",
            "GBP",
            "GEL",
            "GGP",
            "GHS",
            "GIP",
            "GMD",
            "GNF",
            "GTQ",
            "GYD",
            "HKD",
            "HNL",
            "HRK",
            "HTG",
            "HUF",
            "IDR",
            "ILS",
            "IMP",
            "INR",
            "IQD",
            "IRR",
            "ISK",
            "JMD",
            "JOD",
            "JPY",
            "KES",
            "KGS",
            "KHR",
            "KMF",
            "KPW",
            "KRW",
            "KWD",
            "KYD",
            "KZT",
            "LAK",
            "LBP",
            "LKR",
            "LRD",
            "LSL",
            "LYD",
            "MAD",
            "MDL",
            "MGA",
            "MKD",
            "MMK",
            "MNT",
            "MOP",
            "MUR",
            "MVR",
            "MWK",
            "MXN",
            "MYR",
            "MZN",
            "NAD",
            "NGN",
            "NIO",
            "NOK",
            "NPR",
            "NZD",
            "OMR",
            "PAB",
            "PEN",
            "PGK",
            "PHP",
            "PKR",
            "PLN",
            "PYG",
            "QAR",
            "RON",
            "RSD",
            "RUB",
            "RWF",
            "SAR",
            "SBD",
            "SCR",
            "SDG",
            "SEK",
            "SGD",
            "SHP",
            "SLL",
            "SOS",
            "SPL",
            "SRD",
            "SVC",
            "SYP",
            "STN",
            "SZL",
            "THB",
            "TJS",
            "TMT",
            "TND",
            "TOP",
            "TRY",
            "TTD",
            "TVD",
            "TWD",
            "TZS",
            "UAH",
            "UGX",
            "USD",
            "UYU",
            "UZS",
            "VEF",
            "VND",
            "VUV",
            "WST",
            "XAF",
            "XCD",
            "XOF",
            "XPF",
            "YER",
            "ZAR",
            "ZMW",
            "LOGICAL",
            "CUSTOM",
        ]
    ]
    """Currency code (ISO 4217 or crypto)"""

    description: Required[str]
    """Description of the line item"""

    op: Required[Literal["add"]]
    """Operation type - add a new line item"""

    product_id: Required[str]
    """ID of the product/catalog item"""

    type: Required[Literal["payin", "payout"]]
    """The type of the line item"""

    user: Required[LineItemsUpdateAddLineItemOperationUser]

    amount: str
    """Deprecated: use price instead. Total amount in smallest currency unit."""

    price: LineItemsUpdateAddLineItemOperationPrice
    """Price breakdown. Provide amount, or unit_price + quantity, or all three."""

    tags: Iterable[LineItemsUpdateAddLineItemOperationTag]
    """Optional metadata tags for this line item"""


class LineItemsUpdateUpdateLineItemOperationPrice(TypedDict, total=False):
    """Price breakdown. Provide amount, or unit_price + quantity, or all three."""

    amount: str
    """Total amount in smallest currency unit.

    Required if unit_price and quantity are not provided.
    """

    quantity: int
    """Number of units for this line item."""

    unit_price: str
    """Price per unit in smallest currency unit."""


class LineItemsUpdateUpdateLineItemOperation(TypedDict, total=False):
    """Operation to update an existing line item pricing"""

    id: Required[str]
    """ID of the line item to update"""

    op: Required[Literal["update"]]
    """Operation type - update an existing line item"""

    amount: str
    """Deprecated: use price instead. Total amount in smallest currency unit."""

    price: LineItemsUpdateUpdateLineItemOperationPrice
    """Price breakdown. Provide amount, or unit_price + quantity, or all three."""


class LineItemsUpdateDeleteLineItemOperation(TypedDict, total=False):
    """Operation to delete a line item from an invoice"""

    id: Required[str]
    """ID of the line item to delete"""

    op: Required[Literal["delete"]]
    """Operation type - delete an existing line item"""


LineItemsUpdate: TypeAlias = Union[
    LineItemsUpdateAddLineItemOperation, LineItemsUpdateUpdateLineItemOperation, LineItemsUpdateDeleteLineItemOperation
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from fragment import Fragment, AsyncFragment
from tests.utils import assert_matches_type
from fragment.types import (
    InvoiceListResponse,
    InvoiceCreateResponse,
    InvoiceSearchResponse,
    InvoiceUpdateResponse,
    InvoiceRetrieveResponse,
    InvoiceListHistoryResponse,
)

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestInvoices:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_method_create(self, client: Fragment) -> None:
        invoice = client.invoices.create(
            invoice_id="invoice_2024_001",
            line_items=[
                {
                    "description": "Professional services for January 2026",
                    "product_id": "prod_1234567890",
                    "type": "payout",
                    "user": {"id": "user_abc123"},
                }
            ],
        )
        assert_matches_type(InvoiceCreateResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_method_create_with_all_params(self, client: Fragment) -> None:
        invoice = client.invoices.create(
            invoice_id="invoice_2024_001",
            line_items=[
                {
                    "description": "Professional services for January 2026",
                    "product_id": "prod_1234567890",
                    "type": "payout",
                    "user": {"id": "user_abc123"},
                    "amount": "1000",
                    "currency_code": "USD",
                    "price": {
                        "amount": "1000",
                        "quantity": 2,
                        "unit_price": "500",
                    },
                    "tags": [
                        {
                            "key": "region",
                            "value": "us-east",
                        }
                    ],
                }
            ],
            tags=[
                {
                    "key": "region",
                    "value": "us-east",
                }
            ],
        )
        assert_matches_type(InvoiceCreateResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_raw_response_create(self, client: Fragment) -> None:
        response = client.invoices.with_raw_response.create(
            invoice_id="invoice_2024_001",
            line_items=[
                {
                    "description": "Professional services for January 2026",
                    "product_id": "prod_1234567890",
                    "type": "payout",
                    "user": {"id": "user_abc123"},
                }
            ],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(InvoiceCreateResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_streaming_response_create(self, client: Fragment) -> None:
        with client.invoices.with_streaming_response.create(
            invoice_id="invoice_2024_001",
            line_items=[
                {
                    "description": "Professional services for January 2026",
                    "product_id": "prod_1234567890",
                    "type": "payout",
                    "user": {"id": "user_abc123"},
                }
            ],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(InvoiceCreateResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_method_retrieve(self, client: Fragment) -> None:
        invoice = client.invoices.retrieve(
            "inv_1234567890",
        )
        assert_matches_type(InvoiceRetrieveResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_raw_response_retrieve(self, client: Fragment) -> None:
        response = client.invoices.with_raw_response.retrieve(
            "inv_1234567890",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(InvoiceRetrieveResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_streaming_response_retrieve(self, client: Fragment) -> None:
        with client.invoices.with_streaming_response.retrieve(
            "inv_1234567890",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(InvoiceRetrieveResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_path_params_retrieve(self, client: Fragment) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `id` but received ''"):
            client.invoices.with_raw_response.retrieve(
                "",
            )

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_method_update(self, client: Fragment) -> None:
        invoice = client.invoices.update(
            id="inv_1234567890",
            line_items_update=[
                {
                    "currency_code": "USD",
                    "description": "Professional services for January 2026",
                    "op": "add",
                    "product_id": "prod_1234567890",
                    "type": "payout",
                    "user": {"id": "user_abc123"},
                }
            ],
            version=1,
        )
        assert_matches_type(InvoiceUpdateResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_raw_response_update(self, client: Fragment) -> None:
        response = client.invoices.with_raw_response.update(
            id="inv_1234567890",
            line_items_update=[
                {
                    "currency_code": "USD",
                    "description": "Professional services for January 2026",
                    "op": "add",
                    "product_id": "prod_1234567890",
                    "type": "payout",
                    "user": {"id": "user_abc123"},
                }
            ],
            version=1,
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(InvoiceUpdateResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_streaming_response_update(self, client: Fragment) -> None:
        with client.invoices.with_streaming_response.update(
            id="inv_1234567890",
            line_items_update=[
                {
                    "currency_code": "USD",
                    "description": "Professional services for January 2026",
                    "op": "add",
                    "product_id": "prod_1234567890",
                    "type": "payout",
                    "user": {"id": "user_abc123"},
                }
            ],
            version=1,
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(InvoiceUpdateResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_path_params_update(self, client: Fragment) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `id` but received ''"):
            client.invoices.with_raw_response.update(
                id="",
                line_items_update=[
                    {
                        "currency_code": "USD",
                        "description": "Professional services for January 2026",
                        "op": "add",
                        "product_id": "prod_1234567890",
                        "type": "payout",
                        "user": {"id": "user_abc123"},
                    }
                ],
                version=1,
            )

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_method_list(self, client: Fragment) -> None:
        invoice = client.invoices.list()
        assert_matches_type(InvoiceListResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_raw_response_list(self, client: Fragment) -> None:
        response = client.invoices.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(InvoiceListResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_streaming_response_list(self, client: Fragment) -> None:
        with client.invoices.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(InvoiceListResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_method_list_history(self, client: Fragment) -> None:
        invoice = client.invoices.list_history(
            "inv_1234567890",
        )
        assert_matches_type(InvoiceListHistoryResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_raw_response_list_history(self, client: Fragment) -> None:
        response = client.invoices.with_raw_response.list_history(
            "inv_1234567890",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(InvoiceListHistoryResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_streaming_response_list_history(self, client: Fragment) -> None:
        with client.invoices.with_streaming_response.list_history(
            "inv_1234567890",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(InvoiceListHistoryResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_path_params_list_history(self, client: Fragment) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `id` but received ''"):
            client.invoices.with_raw_response.list_history(
                "",
            )

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_method_search(self, client: Fragment) -> None:
        invoice = client.invoices.search(
            filter={},
            page_info={},
        )
        assert_matches_type(InvoiceSearchResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_method_search_with_all_params(self, client: Fragment) -> None:
        invoice = client.invoices.search(
            filter={
                "tags": {
                    "all": [
                        {
                            "key": "env",
                            "value": "prod",
                        }
                    ],
                    "any": [
                        {
                            "key": "region",
                            "value": "us-*",
                        }
                    ],
                }
            },
            page_info={
                "after": "after",
                "limit": 20,
            },
        )
        assert_matches_type(InvoiceSearchResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_raw_response_search(self, client: Fragment) -> None:
        response = client.invoices.with_raw_response.search(
            filter={},
            page_info={},
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = response.parse()
        assert_matches_type(InvoiceSearchResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_streaming_response_search(self, client: Fragment) -> None:
        with client.invoices.with_streaming_response.search(
            filter={},
            page_info={},
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = response.parse()
            assert_matches_type(InvoiceSearchResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True


class TestAsyncInvoices:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_method_create(self, async_client: AsyncFragment) -> None:
        invoice = await async_client.invoices.create(
            invoice_id="invoice_2024_001",
            line_items=[
                {
                    "description": "Professional services for January 2026",
                    "product_id": "prod_1234567890",
                    "type": "payout",
                    "user": {"id": "user_abc123"},
                }
            ],
        )
        assert_matches_type(InvoiceCreateResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_method_create_with_all_params(self, async_client: AsyncFragment) -> None:
        invoice = await async_client.invoices.create(
            invoice_id="invoice_2024_001",
            line_items=[
                {
                    "description": "Professional services for January 2026",
                    "product_id": "prod_1234567890",
                    "type": "payout",
                    "user": {"id": "user_abc123"},
                    "amount": "1000",
                    "currency_code": "USD",
                    "price": {
                        "amount": "1000",
                        "quantity": 2,
                        "unit_price": "500",
                    },
                    "tags": [
                        {
                            "key": "region",
                            "value": "us-east",
                        }
                    ],
                }
            ],
            tags=[
                {
                    "key": "region",
                    "value": "us-east",
                }
            ],
        )
        assert_matches_type(InvoiceCreateResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_raw_response_create(self, async_client: AsyncFragment) -> None:
        response = await async_client.invoices.with_raw_response.create(
            invoice_id="invoice_2024_001",
            line_items=[
                {
                    "description": "Professional services for January 2026",
                    "product_id": "prod_1234567890",
                    "type": "payout",
                    "user": {"id": "user_abc123"},
                }
            ],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = await response.parse()
        assert_matches_type(InvoiceCreateResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_streaming_response_create(self, async_client: AsyncFragment) -> None:
        async with async_client.invoices.with_streaming_response.create(
            invoice_id="invoice_2024_001",
            line_items=[
                {
                    "description": "Professional services for January 2026",
                    "product_id": "prod_1234567890",
                    "type": "payout",
                    "user": {"id": "user_abc123"},
                }
            ],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(InvoiceCreateResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_method_retrieve(self, async_client: AsyncFragment) -> None:
        invoice = await async_client.invoices.retrieve(
            "inv_1234567890",
        )
        assert_matches_type(InvoiceRetrieveResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_raw_response_retrieve(self, async_client: AsyncFragment) -> None:
        response = await async_client.invoices.with_raw_response.retrieve(
            "inv_1234567890",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = await response.parse()
        assert_matches_type(InvoiceRetrieveResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_streaming_response_retrieve(self, async_client: AsyncFragment) -> None:
        async with async_client.invoices.with_streaming_response.retrieve(
            "inv_1234567890",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(InvoiceRetrieveResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_path_params_retrieve(self, async_client: AsyncFragment) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `id` but received ''"):
            await async_client.invoices.with_raw_response.retrieve(
                "",
            )

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_method_update(self, async_client: AsyncFragment) -> None:
        invoice = await async_client.invoices.update(
            id="inv_1234567890",
            line_items_update=[
                {
                    "currency_code": "USD",
                    "description": "Professional services for January 2026",
                    "op": "add",
                    "product_id": "prod_1234567890",
                    "type": "payout",
                    "user": {"id": "user_abc123"},
                }
            ],
            version=1,
        )
        assert_matches_type(InvoiceUpdateResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_raw_response_update(self, async_client: AsyncFragment) -> None:
        response = await async_client.invoices.with_raw_response.update(
            id="inv_1234567890",
            line_items_update=[
                {
                    "currency_code": "USD",
                    "description": "Professional services for January 2026",
                    "op": "add",
                    "product_id": "prod_1234567890",
                    "type": "payout",
                    "user": {"id": "user_abc123"},
                }
            ],
            version=1,
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = await response.parse()
        assert_matches_type(InvoiceUpdateResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_streaming_response_update(self, async_client: AsyncFragment) -> None:
        async with async_client.invoices.with_streaming_response.update(
            id="inv_1234567890",
            line_items_update=[
                {
                    "currency_code": "USD",
                    "description": "Professional services for January 2026",
                    "op": "add",
                    "product_id": "prod_1234567890",
                    "type": "payout",
                    "user": {"id": "user_abc123"},
                }
            ],
            version=1,
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(InvoiceUpdateResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_path_params_update(self, async_client: AsyncFragment) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `id` but received ''"):
            await async_client.invoices.with_raw_response.update(
                id="",
                line_items_update=[
                    {
                        "currency_code": "USD",
                        "description": "Professional services for January 2026",
                        "op": "add",
                        "product_id": "prod_1234567890",
                        "type": "payout",
                        "user": {"id": "user_abc123"},
                    }
                ],
                version=1,
            )

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_method_list(self, async_client: AsyncFragment) -> None:
        invoice = await async_client.invoices.list()
        assert_matches_type(InvoiceListResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_raw_response_list(self, async_client: AsyncFragment) -> None:
        response = await async_client.invoices.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = await response.parse()
        assert_matches_type(InvoiceListResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_streaming_response_list(self, async_client: AsyncFragment) -> None:
        async with async_client.invoices.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(InvoiceListResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_method_list_history(self, async_client: AsyncFragment) -> None:
        invoice = await async_client.invoices.list_history(
            "inv_1234567890",
        )
        assert_matches_type(InvoiceListHistoryResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_raw_response_list_history(self, async_client: AsyncFragment) -> None:
        response = await async_client.invoices.with_raw_response.list_history(
            "inv_1234567890",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = await response.parse()
        assert_matches_type(InvoiceListHistoryResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_streaming_response_list_history(self, async_client: AsyncFragment) -> None:
        async with async_client.invoices.with_streaming_response.list_history(
            "inv_1234567890",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(InvoiceListHistoryResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_path_params_list_history(self, async_client: AsyncFragment) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `id` but received ''"):
            await async_client.invoices.with_raw_response.list_history(
                "",
            )

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_method_search(self, async_client: AsyncFragment) -> None:
        invoice = await async_client.invoices.search(
            filter={},
            page_info={},
        )
        assert_matches_type(InvoiceSearchResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_method_search_with_all_params(self, async_client: AsyncFragment) -> None:
        invoice = await async_client.invoices.search(
            filter={
                "tags": {
                    "all": [
                        {
                            "key": "env",
                            "value": "prod",
                        }
                    ],
                    "any": [
                        {
                            "key": "region",
                            "value": "us-*",
                        }
                    ],
                }
            },
            page_info={
                "after": "after",
                "limit": 20,
            },
        )
        assert_matches_type(InvoiceSearchResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_raw_response_search(self, async_client: AsyncFragment) -> None:
        response = await async_client.invoices.with_raw_response.search(
            filter={},
            page_info={},
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        invoice = await response.parse()
        assert_matches_type(InvoiceSearchResponse, invoice, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_streaming_response_search(self, async_client: AsyncFragment) -> None:
        async with async_client.invoices.with_streaming_response.search(
            filter={},
            page_info={},
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            invoice = await response.parse()
            assert_matches_type(InvoiceSearchResponse, invoice, path=["response"])

        assert cast(Any, response.is_closed) is True



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Iterable
from typing_extensions import Required, TypeAlias, TypedDict

__all__ = [
    "ProductCreateParams",
    "PaidByRole",
    "PaidByRoleRoleMatchByID",
    "PaidByRoleRoleMatchByName",
    "PaidToRole",
    "PaidToRoleRoleMatchByID",
    "PaidToRoleRoleMatchByName",
]


class ProductCreateParams(TypedDict, total=False):
    code: Required[str]
    """Product code (unique identifier)"""

    description: str
    """Description of the product"""

    paid_by_roles: Iterable[PaidByRole]
    """Roles that can pay for this product.

    Reference roles by id or name. At least one of paid_by_roles or paid_to_roles
    must be provided.
    """

    paid_to_roles: Iterable[PaidToRole]
    """Roles that receive payment for this product.

    Reference roles by id or name. At least one of paid_by_roles or paid_to_roles
    must be provided.
    """


class PaidByRoleRoleMatchByID(TypedDict, total=False):
    id: Required[str]
    """The unique ID of the role"""


class PaidByRoleRoleMatchByName(TypedDict, total=False):
    name: Required[str]
    """The name of the role"""


PaidByRole: TypeAlias = Union[PaidByRoleRoleMatchByID, PaidByRoleRoleMatchByName]


class PaidToRoleRoleMatchByID(TypedDict, total=False):
    id: Required[str]
    """The unique ID of the role"""


class PaidToRoleRoleMatchByName(TypedDict, total=False):
    name: Required[str]
    """The name of the role"""


PaidToRole: TypeAlias = Union[PaidToRoleRoleMatchByID, PaidToRoleRoleMatchByName]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .._models import BaseModel

__all__ = ["User"]


class User(BaseModel):
    """User object"""

    id: str
    """Unique user ID"""

    external_id: str
    """External ID for the user"""

    role: str
    """Role of the user"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from fragment import Fragment, AsyncFragment
from tests.utils import assert_matches_type
from fragment.types import ProductListResponse, ProductCreateResponse, ProductRetrieveResponse

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestProducts:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_method_create(self, client: Fragment) -> None:
        product = client.products.create(
            code="PROD_001",
        )
        assert_matches_type(ProductCreateResponse, product, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_method_create_with_all_params(self, client: Fragment) -> None:
        product = client.products.create(
            code="PROD_001",
            description="Premium subscription service",
            paid_by_roles=[{"name": "buyer"}],
            paid_to_roles=[{"name": "seller"}],
        )
        assert_matches_type(ProductCreateResponse, product, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_raw_response_create(self, client: Fragment) -> None:
        response = client.products.with_raw_response.create(
            code="PROD_001",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        product = response.parse()
        assert_matches_type(ProductCreateResponse, product, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_streaming_response_create(self, client: Fragment) -> None:
        with client.products.with_streaming_response.create(
            code="PROD_001",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            product = response.parse()
            assert_matches_type(ProductCreateResponse, product, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_method_retrieve(self, client: Fragment) -> None:
        product = client.products.retrieve(
            "PROD_001",
        )
        assert_matches_type(ProductRetrieveResponse, product, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_raw_response_retrieve(self, client: Fragment) -> None:
        response = client.products.with_raw_response.retrieve(
            "PROD_001",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        product = response.parse()
        assert_matches_type(ProductRetrieveResponse, product, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_streaming_response_retrieve(self, client: Fragment) -> None:
        with client.products.with_streaming_response.retrieve(
            "PROD_001",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            product = response.parse()
            assert_matches_type(ProductRetrieveResponse, product, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_path_params_retrieve(self, client: Fragment) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `code` but received ''"):
            client.products.with_raw_response.retrieve(
                "",
            )

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_method_list(self, client: Fragment) -> None:
        product = client.products.list()
        assert_matches_type(ProductListResponse, product, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_raw_response_list(self, client: Fragment) -> None:
        response = client.products.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        product = response.parse()
        assert_matches_type(ProductListResponse, product, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    def test_streaming_response_list(self, client: Fragment) -> None:
        with client.products.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            product = response.parse()
            assert_matches_type(ProductListResponse, product, path=["response"])

        assert cast(Any, response.is_closed) is True


class TestAsyncProducts:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_method_create(self, async_client: AsyncFragment) -> None:
        product = await async_client.products.create(
            code="PROD_001",
        )
        assert_matches_type(ProductCreateResponse, product, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_method_create_with_all_params(self, async_client: AsyncFragment) -> None:
        product = await async_client.products.create(
            code="PROD_001",
            description="Premium subscription service",
            paid_by_roles=[{"name": "buyer"}],
            paid_to_roles=[{"name": "seller"}],
        )
        assert_matches_type(ProductCreateResponse, product, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_raw_response_create(self, async_client: AsyncFragment) -> None:
        response = await async_client.products.with_raw_response.create(
            code="PROD_001",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        product = await response.parse()
        assert_matches_type(ProductCreateResponse, product, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_streaming_response_create(self, async_client: AsyncFragment) -> None:
        async with async_client.products.with_streaming_response.create(
            code="PROD_001",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            product = await response.parse()
            assert_matches_type(ProductCreateResponse, product, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_method_retrieve(self, async_client: AsyncFragment) -> None:
        product = await async_client.products.retrieve(
            "PROD_001",
        )
        assert_matches_type(ProductRetrieveResponse, product, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_raw_response_retrieve(self, async_client: AsyncFragment) -> None:
        response = await async_client.products.with_raw_response.retrieve(
            "PROD_001",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        product = await response.parse()
        assert_matches_type(ProductRetrieveResponse, product, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_streaming_response_retrieve(self, async_client: AsyncFragment) -> None:
        async with async_client.products.with_streaming_response.retrieve(
            "PROD_001",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            product = await response.parse()
            assert_matches_type(ProductRetrieveResponse, product, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_path_params_retrieve(self, async_client: AsyncFragment) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `code` but received ''"):
            await async_client.products.with_raw_response.retrieve(
                "",
            )

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_method_list(self, async_client: AsyncFragment) -> None:
        product = await async_client.products.list()
        assert_matches_type(ProductListResponse, product, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_raw_response_list(self, async_client: AsyncFragment) -> None:
        response = await async_client.products.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        product = await response.parse()
        assert_matches_type(ProductListResponse, product, path=["response"])

    @pytest.mark.skip(reason="Mock server tests are disabled")
    @parametrize
    async def test_streaming_response_list(self, async_client: AsyncFragment) -> None:
        async with async_client.products.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            product = await response.parse()
            assert_matches_type(ProductListResponse, product, path=["response"])

        assert cast(Any, response.is_closed) is True



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List

from .._models import BaseModel
from .transaction import Transaction

__all__ = ["TransactionListResponse"]


class TransactionListResponse(BaseModel):
    """List of transactions"""

    data: List[Transaction]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List

from .product import Product
from .._models import BaseModel

__all__ = ["ProductListResponse"]


class ProductListResponse(BaseModel):
    """List of products"""

    data: List[Product]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Mapping
from typing_extensions import Self, override

import httpx

from . import _exceptions
from ._qs import Querystring
from ._types import (
    Omit,
    Timeout,
    NotGiven,
    Transport,
    ProxiesTypes,
    RequestOptions,
    not_given,
)
from ._utils import is_given, get_async_library
from ._compat import cached_property
from ._oauth2 import OAuth2ClientCredentials, make_oauth2
from ._version import __version__
from ._streaming import Stream as Stream, AsyncStream as AsyncStream
from ._exceptions import APIStatusError
from ._base_client import (
    DEFAULT_MAX_RETRIES,
    SyncAPIClient,
    AsyncAPIClient,
)

if TYPE_CHECKING:
    from .resources import roles, users, invoices, products, transactions, external_accounts
    from .resources.roles import RolesResource, AsyncRolesResource
    from .resources.users import UsersResource, AsyncUsersResource
    from .resources.invoices import InvoicesResource, AsyncInvoicesResource
    from .resources.products import ProductsResource, AsyncProductsResource
    from .resources.transactions import TransactionsResource, AsyncTransactionsResource
    from .resources.external_accounts import ExternalAccountsResource, AsyncExternalAccountsResource

__all__ = [
    "Timeout",
    "Transport",
    "ProxiesTypes",
    "RequestOptions",
    "Fragment",
    "AsyncFragment",
    "Client",
    "AsyncClient",
]


class Fragment(SyncAPIClient):
    # client options
    client_id: str | None
    client_secret: str | None

    def __init__(
        self,
        *,
        client_id: str | None = None,
        client_secret: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client.
        # We provide a `DefaultHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#client) for more details.
        http_client: httpx.Client | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new synchronous Fragment client instance.

        This automatically infers the following arguments from their corresponding environment variables if they are not provided:
        - `client_id` from `FRAGMENT_CLIENT_ID`
        - `client_secret` from `FRAGMENT_CLIENT_SECRET`
        """
        if client_id is None:
            client_id = os.environ.get("FRAGMENT_CLIENT_ID")
        self.client_id = client_id

        if client_secret is None:
            client_secret = os.environ.get("FRAGMENT_CLIENT_SECRET")
        self.client_secret = client_secret

        if base_url is None:
            base_url = os.environ.get("FRAGMENT_BASE_URL")
        if base_url is None:
            base_url = f"https://api.us-west-2.fragment.dev"

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

    @cached_property
    def external_accounts(self) -> ExternalAccountsResource:
        from .resources.external_accounts import ExternalAccountsResource

        return ExternalAccountsResource(self)

    @cached_property
    def invoices(self) -> InvoicesResource:
        """Invoice management operations"""
        from .resources.invoices import InvoicesResource

        return InvoicesResource(self)

    @cached_property
    def products(self) -> ProductsResource:
        """Product management operations"""
        from .resources.products import ProductsResource

        return ProductsResource(self)

    @cached_property
    def roles(self) -> RolesResource:
        """Role management operations"""
        from .resources.roles import RolesResource

        return RolesResource(self)

    @cached_property
    def transactions(self) -> TransactionsResource:
        """Transaction sync operations"""
        from .resources.transactions import TransactionsResource

        return TransactionsResource(self)

    @cached_property
    def users(self) -> UsersResource:
        from .resources.users import UsersResource

        return UsersResource(self)

    @cached_property
    def with_raw_response(self) -> FragmentWithRawResponse:
        return FragmentWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> FragmentWithStreamedResponse:
        return FragmentWithStreamedResponse(self)

    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="comma")

    @property
    @override
    def custom_auth(self) -> httpx.Auth | None:
        if self.client_id and self.client_secret:
            return make_oauth2(
                client_id=self.client_id,
                client_secret=self.client_secret,
                token_url=self._prepare_url("https://auth.us-west-2.fragment.dev/oauth2/token"),
                header="Authorization",
            )
        return None

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": "false",
            **self._custom_headers,
        }

    @override
    def _should_retry(self, response: httpx.Response) -> bool:
        # Retry on 401 if we are using OAuth2 and the token might be expired
        if response.status_code == 401 and isinstance(self.custom_auth, OAuth2ClientCredentials):
            if self.custom_auth.token_is_expired():
                self.custom_auth.invalidate_token()
                return True
        return super()._should_retry(response)

    def copy(
        self,
        *,
        client_id: str | None = None,
        client_secret: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        http_client: httpx.Client | None = None,
        max_retries: int | NotGiven = not_given,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")

        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            client_id=client_id or self.client_id,
            client_secret=client_secret or self.client_secret,
            base_url=base_url or self.base_url,
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
            **_extra_kwargs,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(err_msg, response=response, body=body)

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)


class AsyncFragment(AsyncAPIClient):
    # client options
    client_id: str | None
    client_secret: str | None

    def __init__(
        self,
        *,
        client_id: str | None = None,
        client_secret: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client.
        # We provide a `DefaultAsyncHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#asyncclient) for more details.
        http_client: httpx.AsyncClient | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new async AsyncFragment client instance.

        This automatically infers the following arguments from their corresponding environment variables if they are not provided:
        - `client_id` from `FRAGMENT_CLIENT_ID`
        - `client_secret` from `FRAGMENT_CLIENT_SECRET`
        """
        if client_id is None:
            client_id = os.environ.get("FRAGMENT_CLIENT_ID")
        self.client_id = client_id

        if client_secret is None:
            client_secret = os.environ.get("FRAGMENT_CLIENT_SECRET")
        self.client_secret = client_secret

        if base_url is None:
            base_url = os.environ.get("FRAGMENT_BASE_URL")
        if base_url is None:
            base_url = f"https://api.us-west-2.fragment.dev"

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

    @cached_property
    def external_accounts(self) -> AsyncExternalAccountsResource:
        from .resources.external_accounts import AsyncExternalAccountsResource

        return AsyncExternalAccountsResource(self)

    @cached_property
    def invoices(self) -> AsyncInvoicesResource:
        """Invoice management operations"""
        from .resources.invoices import AsyncInvoicesResource

        return AsyncInvoicesResource(self)

    @cached_property
    def products(self) -> AsyncProductsResource:
        """Product management operations"""
        from .resources.products import AsyncProductsResource

        return AsyncProductsResource(self)

    @cached_property
    def roles(self) -> AsyncRolesResource:
        """Role management operations"""
        from .resources.roles import AsyncRolesResource

        return AsyncRolesResource(self)

    @cached_property
    def transactions(self) -> AsyncTransactionsResource:
        """Transaction sync operations"""
        from .resources.transactions import AsyncTransactionsResource

        return AsyncTransactionsResource(self)

    @cached_property
    def users(self) -> AsyncUsersResource:
        from .resources.users import AsyncUsersResource

        return AsyncUsersResource(self)

    @cached_property
    def with_raw_response(self) -> AsyncFragmentWithRawResponse:
        return AsyncFragmentWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncFragmentWithStreamedResponse:
        return AsyncFragmentWithStreamedResponse(self)

    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="comma")

    @property
    @override
    def custom_auth(self) -> httpx.Auth | None:
        if self.client_id and self.client_secret:
            return make_oauth2(
                client_id=self.client_id,
                client_secret=self.client_secret,
                token_url=self._prepare_url("https://auth.us-west-2.fragment.dev/oauth2/token"),
                header="Authorization",
            )
        return None

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": f"async:{get_async_library()}",
            **self._custom_headers,
        }

    @override
    def _should_retry(self, response: httpx.Response) -> bool:
        # Retry on 401 if we are using OAuth2 and the token might be expired
        if response.status_code == 401 and isinstance(self.custom_auth, OAuth2ClientCredentials):
            if self.custom_auth.token_is_expired():
                self.custom_auth.invalidate_token()
                return True
        return super()._should_retry(response)

    def copy(
        self,
        *,
        client_id: str | None = None,
        client_secret: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        http_client: httpx.AsyncClient | None = None,
        max_retries: int | NotGiven = not_given,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")

        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            client_id=client_id or self.client_id,
            client_secret=client_secret or self.client_secret,
            base_url=base_url or self.base_url,
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
            **_extra_kwargs,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(err_msg, response=response, body=body)

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)


class FragmentWithRawResponse:
    _client: Fragment

    def __init__(self, client: Fragment) -> None:
        self._client = client

    @cached_property
    def external_accounts(self) -> external_accounts.ExternalAccountsResourceWithRawResponse:
        from .resources.external_accounts import ExternalAccountsResourceWithRawResponse

        return ExternalAccountsResourceWithRawResponse(self._client.external_accounts)

    @cached_property
    def invoices(self) -> invoices.InvoicesResourceWithRawResponse:
        """Invoice management operations"""
        from .resources.invoices import InvoicesResourceWithRawResponse

        return InvoicesResourceWithRawResponse(self._client.invoices)

    @cached_property
    def products(self) -> products.ProductsResourceWithRawResponse:
        """Product management operations"""
        from .resources.products import ProductsResourceWithRawResponse

        return ProductsResourceWithRawResponse(self._client.products)

    @cached_property
    def roles(self) -> roles.RolesResourceWithRawResponse:
        """Role management operations"""
        from .resources.roles import RolesResourceWithRawResponse

        return RolesResourceWithRawResponse(self._client.roles)

    @cached_property
    def transactions(self) -> transactions.TransactionsResourceWithRawResponse:
        """Transaction sync operations"""
        from .resources.transactions import TransactionsResourceWithRawResponse

        return TransactionsResourceWithRawResponse(self._client.transactions)

    @cached_property
    def users(self) -> users.UsersResourceWithRawResponse:
        from .resources.users import UsersResourceWithRawResponse

        return UsersResourceWithRawResponse(self._client.users)


class AsyncFragmentWithRawResponse:
    _client: AsyncFragment

    def __init__(self, client: AsyncFragment) -> None:
        self._client = client

    @cached_property
    def external_accounts(self) -> external_accounts.AsyncExternalAccountsResourceWithRawResponse:
        from .resources.external_accounts import AsyncExternalAccountsResourceWithRawResponse

        return AsyncExternalAccountsResourceWithRawResponse(self._client.external_accounts)

    @cached_property
    def invoices(self) -> invoices.AsyncInvoicesResourceWithRawResponse:
        """Invoice management operations"""
        from .resources.invoices import AsyncInvoicesResourceWithRawResponse

        return AsyncInvoicesResourceWithRawResponse(self._client.invoices)

    @cached_property
    def products(self) -> products.AsyncProductsResourceWithRawResponse:
        """Product management operations"""
        from .resources.products import AsyncProductsResourceWithRawResponse

        return AsyncProductsResourceWithRawResponse(self._client.products)

    @cached_property
    def roles(self) -> roles.AsyncRolesResourceWithRawResponse:
        """Role management operations"""
        from .resources.roles import AsyncRolesResourceWithRawResponse

        return AsyncRolesResourceWithRawResponse(self._client.roles)

    @cached_property
    def transactions(self) -> transactions.AsyncTransactionsResourceWithRawResponse:
        """Transaction sync operations"""
        from .resources.transactions import AsyncTransactionsResourceWithRawResponse

        return AsyncTransactionsResourceWithRawResponse(self._client.transactions)

    @cached_property
    def users(self) -> users.AsyncUsersResourceWithRawResponse:
        from .resources.users import AsyncUsersResourceWithRawResponse

        return AsyncUsersResourceWithRawResponse(self._client.users)


class FragmentWithStreamedResponse:
    _client: Fragment

    def __init__(self, client: Fragment) -> None:
        self._client = client

    @cached_property
    def external_accounts(self) -> external_accounts.ExternalAccountsResourceWithStreamingResponse:
        from .resources.external_accounts import ExternalAccountsResourceWithStreamingResponse

        return ExternalAccountsResourceWithStreamingResponse(self._client.external_accounts)

    @cached_property
    def invoices(self) -> invoices.InvoicesResourceWithStreamingResponse:
        """Invoice management operations"""
        from .resources.invoices import InvoicesResourceWithStreamingResponse

        return InvoicesResourceWithStreamingResponse(self._client.invoices)

    @cached_property
    def products(self) -> products.ProductsResourceWithStreamingResponse:
        """Product management operations"""
        from .resources.products import ProductsResourceWithStreamingResponse

        return ProductsResourceWithStreamingResponse(self._client.products)

    @cached_property
    def roles(self) -> roles.RolesResourceWithStreamingResponse:
        """Role management operations"""
        from .resources.roles import RolesResourceWithStreamingResponse

        return RolesResourceWithStreamingResponse(self._client.roles)

    @cached_property
    def transactions(self) -> transactions.TransactionsResourceWithStreamingResponse:
        """Transaction sync operations"""
        from .resources.transactions import TransactionsResourceWithStreamingResponse

        return TransactionsResourceWithStreamingResponse(self._client.transactions)

    @cached_property
    def users(self) -> users.UsersResourceWithStreamingResponse:
        from .resources.users import UsersResourceWithStreamingResponse

        return UsersResourceWithStreamingResponse(self._client.users)


class AsyncFragmentWithStreamedResponse:
    _client: AsyncFragment

    def __init__(self, client: AsyncFragment) -> None:
        self._client = client

    @cached_property
    def external_accounts(self) -> external_accounts.AsyncExternalAccountsResourceWithStreamingResponse:
        from .resources.external_accounts import AsyncExternalAccountsResourceWithStreamingResponse

        return AsyncExternalAccountsResourceWithStreamingResponse(self._client.external_accounts)

    @cached_property
    def invoices(self) -> invoices.AsyncInvoicesResourceWithStreamingResponse:
        """Invoice management operations"""
        from .resources.invoices import AsyncInvoicesResourceWithStreamingResponse

        return AsyncInvoicesResourceWithStreamingResponse(self._client.invoices)

    @cached_property
    def products(self) -> products.AsyncProductsResourceWithStreamingResponse:
        """Product management operations"""
        from .resources.products import AsyncProductsResourceWithStreamingResponse

        return AsyncProductsResourceWithStreamingResponse(self._client.products)

    @cached_property
    def roles(self) -> roles.AsyncRolesResourceWithStreamingResponse:
        """Role management operations"""
        from .resources.roles import AsyncRolesResourceWithStreamingResponse

        return AsyncRolesResourceWithStreamingResponse(self._client.roles)

    @cached_property
    def transactions(self) -> transactions.AsyncTransactionsResourceWithStreamingResponse:
        """Transaction sync operations"""
        from .resources.transactions import AsyncTransactionsResourceWithStreamingResponse

        return AsyncTransactionsResourceWithStreamingResponse(self._client.transactions)

    @cached_property
    def users(self) -> users.AsyncUsersResourceWithStreamingResponse:
        from .resources.users import AsyncUsersResourceWithStreamingResponse

        return AsyncUsersResourceWithStreamingResponse(self._client.users)


Client = Fragment

AsyncClient = AsyncFragment



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List

from .role import Role
from .._models import BaseModel

__all__ = ["RoleListResponse"]


class RoleListResponse(BaseModel):
    """List of roles"""

    data: List[Role]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable
from typing_extensions import Required, TypedDict

__all__ = ["InvoiceSearchParams", "Filter", "FilterTags", "FilterTagsAll", "FilterTagsAny", "PageInfo"]


class InvoiceSearchParams(TypedDict, total=False):
    filter: Required[Filter]
    """Filter criteria for the search"""

    page_info: Required[PageInfo]
    """Pagination parameters"""


class FilterTagsAll(TypedDict, total=False):
    """A tag filter."""

    key: Required[str]
    """Tag key to filter on. Must be an exact match; wildcards are not supported."""

    value: Required[str]
    """Tag value pattern to filter on.

    Supports wildcards: `*` matches any characters, `?` matches a single character.
    Use `\\**` or `\\??` to match literal asterisks or question marks. Use `*` to match
    any value for the given key.
    """


class FilterTagsAny(TypedDict, total=False):
    """A tag filter."""

    key: Required[str]
    """Tag key to filter on. Must be an exact match; wildcards are not supported."""

    value: Required[str]
    """Tag value pattern to filter on.

    Supports wildcards: `*` matches any characters, `?` matches a single character.
    Use `\\**` or `\\??` to match literal asterisks or question marks. Use `*` to match
    any value for the given key.
    """


class FilterTags(TypedDict, total=False):
    """Tag-based filter criteria.

    When both `any` and `all` are provided, results must match every entry in `all` AND at least one entry in `any`.
    """

    all: Iterable[FilterTagsAll]
    """Returns invoices matching every specified tag (AND)."""

    any: Iterable[FilterTagsAny]
    """Returns invoices matching at least one of the specified tags (OR)."""


class Filter(TypedDict, total=False):
    """Filter criteria for the search"""

    tags: FilterTags
    """Tag-based filter criteria.

    When both `any` and `all` are provided, results must match every entry in `all`
    AND at least one entry in `any`.
    """


class PageInfo(TypedDict, total=False):
    """Pagination parameters"""

    after: str
    """Cursor for fetching the next page of results"""

    limit: int
    """Number of results to return. Defaults to 20."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .role import Role
from .._models import BaseModel

__all__ = ["RoleCreateResponse"]


class RoleCreateResponse(BaseModel):
    data: Role
    """Role object"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List

from .invoice import Invoice
from .._models import BaseModel

__all__ = ["InvoiceListResponse"]


class InvoiceListResponse(BaseModel):
    """List of invoices"""

    data: List[Invoice]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .._models import BaseModel

__all__ = ["ExternalAccount"]


class ExternalAccount(BaseModel):
    """External account object"""

    id: str
    """Fragment-generated unique ID for the external account"""

    external_id: str
    """User-provided external ID"""

    name: str
    """Human-readable name for the external account"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Literal, TypedDict

__all__ = ["TransactionListParams"]


class TransactionListParams(TypedDict, total=False):
    account: str
    """Filter by account.

    Encoded account ID (ext_account_xxx) or external_id. If the account does not
    exist, returns an empty list.
    """

    reconciliation_status: Literal["reconciled", "unreconciled"]
    """Filter by reconciliation state.

    reconciled = unallocated_amount === 0; unreconciled = unallocated_amount !== 0.
    Omit for all transactions.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .invoice import Invoice
from .._models import BaseModel

__all__ = ["InvoiceCreateResponse"]


class InvoiceCreateResponse(BaseModel):
    data: Invoice
    """Invoice object"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from datetime import datetime
from typing_extensions import Literal

from .._models import BaseModel

__all__ = ["TransactionSearchAllocationsResponse", "Data", "DataTransaction", "DataUser"]


class DataTransaction(BaseModel):
    """Reference to the parent transaction."""

    id: str
    """Encoded transaction ID."""

    external_id: str
    """External transaction ID."""


class DataUser(BaseModel):
    id: str
    """FRAGMENT generated ID of the user"""

    external_id: Optional[str] = None
    """External ID of the user"""


class Data(BaseModel):
    """A flattened allocation with a reference to its parent transaction."""

    id: str
    """Allocation ID."""

    amount: str
    """Amount to allocate in smallest currency unit as stringified bigint."""

    invoice_id: str
    """The invoice to allocate against."""

    posted: datetime
    """Posted timestamp of the parent transaction in ISO 8601 format."""

    transaction: DataTransaction
    """Reference to the parent transaction."""

    type: Literal["invoice_payin", "invoice_payout"]
    """The type of allocation."""

    user: DataUser


class TransactionSearchAllocationsResponse(BaseModel):
    """Search results for transaction allocations."""

    data: List[Data]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from datetime import datetime
from typing_extensions import Literal

from .._models import BaseModel

__all__ = ["Transaction", "Account", "Allocation", "AllocationUser", "Tag"]


class Account(BaseModel):
    id: str
    """User-facing encoded account ID."""

    external_id: str
    """External account reference ID."""


class AllocationUser(BaseModel):
    id: str
    """FRAGMENT generated ID of the user"""

    external_id: Optional[str] = None
    """External ID of the user"""


class Allocation(BaseModel):
    """Transaction allocation against an invoice."""

    amount: str
    """Amount to allocate in smallest currency unit as stringified bigint."""

    invoice_id: str
    """The invoice to allocate against."""

    type: Literal["invoice_payin", "invoice_payout"]
    """The type of allocation."""

    user: AllocationUser


class Tag(BaseModel):
    """A key-value tag pair"""

    key: str
    """Tag key"""

    value: str
    """Tag value"""


class Transaction(BaseModel):
    """Transaction object."""

    id: str
    """User-facing encoded transaction ID."""

    account: Account

    allocations: List[Allocation]

    amount: str
    """
    Amount in smallest currency unit as stringified bigint (can be positive or
    negative).
    """

    created: datetime
    """Creation timestamp."""

    currency: Literal[
        "ADA",
        "BTC",
        "DAI",
        "ETH",
        "SOL",
        "USDC",
        "USDT",
        "USDG",
        "EURC",
        "CADC",
        "CADT",
        "XLM",
        "UNI",
        "BCH",
        "LTC",
        "AAVE",
        "LINK",
        "MATIC",
        "PTS",
        "AED",
        "AFN",
        "ALL",
        "AMD",
        "ANG",
        "AOA",
        "ARS",
        "AUD",
        "AWG",
        "AZN",
        "BAM",
        "BBD",
        "BDT",
        "BGN",
        "BHD",
        "BIF",
        "BMD",
        "BND",
        "BOB",
        "BRL",
        "BSD",
        "BTN",
        "BWP",
        "BYR",
        "BZD",
        "CAD",
        "CDF",
        "CHF",
        "CLP",
        "CNY",
        "COP",
        "CRC",
        "CUC",
        "CUP",
        "CVE",
        "CZK",
        "DJF",
        "DKK",
        "DOP",
        "DZD",
        "EGP",
        "ERN",
        "ETB",
        "EUR",
        "FJD",
        "FKP",
        "GBP",
        "GEL",
        "GGP",
        "GHS",
        "GIP",
        "GMD",
        "GNF",
        "GTQ",
        "GYD",
        "HKD",
        "HNL",
        "HRK",
        "HTG",
        "HUF",
        "IDR",
        "ILS",
        "IMP",
        "INR",
        "IQD",
        "IRR",
        "ISK",
        "JMD",
        "JOD",
        "JPY",
        "KES",
        "KGS",
        "KHR",
        "KMF",
        "KPW",
        "KRW",
        "KWD",
        "KYD",
        "KZT",
        "LAK",
        "LBP",
        "LKR",
        "LRD",
        "LSL",
        "LYD",
        "MAD",
        "MDL",
        "MGA",
        "MKD",
        "MMK",
        "MNT",
        "MOP",
        "MUR",
        "MVR",
        "MWK",
        "MXN",
        "MYR",
        "MZN",
        "NAD",
        "NGN",
        "NIO",
        "NOK",
        "NPR",
        "NZD",
        "OMR",
        "PAB",
        "PEN",
        "PGK",
        "PHP",
        "PKR",
        "PLN",
        "PYG",
        "QAR",
        "RON",
        "RSD",
        "RUB",
        "RWF",
        "SAR",
        "SBD",
        "SCR",
        "SDG",
        "SEK",
        "SGD",
        "SHP",
        "SLL",
        "SOS",
        "SPL",
        "SRD",
        "SVC",
        "SYP",
        "STN",
        "SZL",
        "THB",
        "TJS",
        "TMT",
        "TND",
        "TOP",
        "TRY",
        "TTD",
        "TVD",
        "TWD",
        "TZS",
        "UAH",
        "UGX",
        "USD",
        "UYU",
        "UZS",
        "VEF",
        "VND",
        "VUV",
        "WST",
        "XAF",
        "XCD",
        "XOF",
        "XPF",
        "YER",
        "ZAR",
        "ZMW",
        "LOGICAL",
        "CUSTOM",
    ]
    """Currency code (ISO 4217 or crypto)"""

    external_id: str
    """External idempotency key provided by the user."""

    posted: datetime
    """Posted timestamp in ISO 8601 format."""

    tags: List[Tag]
    """Metadata tags associated with this transaction."""

    unallocated_amount: str
    """Read-only amount not yet allocated."""

    version: int
    """Current version of the transaction, used for optimistic concurrency control."""

    modified: Optional[datetime] = None
    """Last modified timestamp."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, TypedDict

__all__ = ["UserCreateParams"]


class UserCreateParams(TypedDict, total=False):
    external_id: Required[str]
    """External ID for the user"""

    role: Required[str]
    """Role of the user"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List

from .user import User
from .._models import BaseModel

__all__ = ["UserListResponse"]


class UserListResponse(BaseModel):
    """List of users"""

    data: List[User]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from .role import Role as Role
from .user import User as User
from .invoice import Invoice as Invoice
from .product import Product as Product
from .transaction import Transaction as Transaction
from .external_account import ExternalAccount as ExternalAccount
from .role_create_params import RoleCreateParams as RoleCreateParams
from .role_list_response import RoleListResponse as RoleListResponse
from .user_create_params import UserCreateParams as UserCreateParams
from .user_list_response import UserListResponse as UserListResponse
from .role_create_response import RoleCreateResponse as RoleCreateResponse
from .user_create_response import UserCreateResponse as UserCreateResponse
from .invoice_create_params import InvoiceCreateParams as InvoiceCreateParams
from .invoice_list_response import InvoiceListResponse as InvoiceListResponse
from .invoice_search_params import InvoiceSearchParams as InvoiceSearchParams
from .invoice_update_params import InvoiceUpdateParams as InvoiceUpdateParams
from .product_create_params import ProductCreateParams as ProductCreateParams
from .product_list_response import ProductListResponse as ProductListResponse
from .invoice_create_response import InvoiceCreateResponse as InvoiceCreateResponse
from .invoice_search_response import InvoiceSearchResponse as InvoiceSearchResponse
from .invoice_update_response import InvoiceUpdateResponse as InvoiceUpdateResponse
from .product_create_response import ProductCreateResponse as ProductCreateResponse
from .transaction_list_params import TransactionListParams as TransactionListParams
from .invoice_retrieve_response import InvoiceRetrieveResponse as InvoiceRetrieveResponse
from .product_retrieve_response import ProductRetrieveResponse as ProductRetrieveResponse
from .transaction_create_params import TransactionCreateParams as TransactionCreateParams
from .transaction_list_response import TransactionListResponse as TransactionListResponse
from .transaction_search_params import TransactionSearchParams as TransactionSearchParams
from .transaction_create_response import TransactionCreateResponse as TransactionCreateResponse
from .transaction_search_response import TransactionSearchResponse as TransactionSearchResponse
from .invoice_list_history_response import InvoiceListHistoryResponse as InvoiceListHistoryResponse
from .transaction_retrieve_response import TransactionRetrieveResponse as TransactionRetrieveResponse
from .external_account_create_params import ExternalAccountCreateParams as ExternalAccountCreateParams
from .external_account_list_response import ExternalAccountListResponse as ExternalAccountListResponse
from .external_account_create_response import ExternalAccountCreateResponse as ExternalAccountCreateResponse
from .transaction_list_history_response import TransactionListHistoryResponse as TransactionListHistoryResponse
from .transaction_create_allocations_params import (
    TransactionCreateAllocationsParams as TransactionCreateAllocationsParams,
)
from .transaction_search_allocations_params import (
    TransactionSearchAllocationsParams as TransactionSearchAllocationsParams,
)
from .transaction_create_allocations_response import (
    TransactionCreateAllocationsResponse as TransactionCreateAllocationsResponse,
)
from .transaction_search_allocations_response import (
    TransactionSearchAllocationsResponse as TransactionSearchAllocationsResponse,
)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from datetime import datetime

from .._models import BaseModel

__all__ = ["Product", "PaidByRole", "PaidToRole"]


class PaidByRole(BaseModel):
    """Reference to a role by its unique ID"""

    id: str
    """The unique ID of the role"""

    name: str
    """The name of the role"""


class PaidToRole(BaseModel):
    """Reference to a role by its unique ID"""

    id: str
    """The unique ID of the role"""

    name: str
    """The name of the role"""


class Product(BaseModel):
    """Product object"""

    id: str
    """Unique identifier for the product"""

    code: str
    """User-defined product identifier."""

    created: datetime
    """ISO 8601 timestamp when the product was created"""

    paid_by_roles: List[PaidByRole]
    """User roles that can pay for this product"""

    paid_to_roles: List[PaidToRole]
    """User roles that receive payment for this product"""

    update_version: float
    """Version number for optimistic locking"""

    workspace_id: str
    """Workspace ID this product belongs to"""

    description: Optional[str] = None
    """Description of the product"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable

import httpx

from ..types import invoice_create_params, invoice_search_params, invoice_update_params
from .._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from .._utils import path_template, maybe_transform, async_maybe_transform
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from .._base_client import make_request_options
from ..types.invoice_list_response import InvoiceListResponse
from ..types.invoice_create_response import InvoiceCreateResponse
from ..types.invoice_search_response import InvoiceSearchResponse
from ..types.invoice_update_response import InvoiceUpdateResponse
from ..types.invoice_retrieve_response import InvoiceRetrieveResponse
from ..types.invoice_list_history_response import InvoiceListHistoryResponse

__all__ = ["InvoicesResource", "AsyncInvoicesResource"]


class InvoicesResource(SyncAPIResource):
    """Invoice management operations"""

    @cached_property
    def with_raw_response(self) -> InvoicesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/fragment-dev/fragment-py#accessing-raw-response-data-eg-headers
        """
        return InvoicesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> InvoicesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/fragment-dev/fragment-py#with_streaming_response
        """
        return InvoicesResourceWithStreamingResponse(self)

    def create(
        self,
        *,
        invoice_id: str,
        line_items: Iterable[invoice_create_params.LineItem],
        tags: Iterable[invoice_create_params.Tag] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceCreateResponse:
        """Creates a new invoice

        Args:
          invoice_id: Unique identifier for the invoice.

        Make this the canonical ID from your system
              for the transaction.

          line_items: List of line items to create with the invoice

          tags: Optional metadata tags for this invoice

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/invoices",
            body=maybe_transform(
                {
                    "invoice_id": invoice_id,
                    "line_items": line_items,
                    "tags": tags,
                },
                invoice_create_params.InvoiceCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=InvoiceCreateResponse,
        )

    def retrieve(
        self,
        id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceRetrieveResponse:
        """
        Gets an invoice by ID with balance details

        Args:
          id: Invoice ID

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return self._get(
            path_template("/invoices/{id}", id=id),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=InvoiceRetrieveResponse,
        )

    def update(
        self,
        id: str,
        *,
        line_items_update: Iterable[invoice_update_params.LineItemsUpdate],
        version: float,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceUpdateResponse:
        """
        Updates an invoice with line item operations

        Args:
          id: Invoice ID

          line_items_update: List of line item operations to apply to the invoice

          version: The version of the invoice being updated. Must match the current version for the
              update to succeed.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return self._post(
            path_template("/invoices/{id}", id=id),
            body=maybe_transform(
                {
                    "line_items_update": line_items_update,
                    "version": version,
                },
                invoice_update_params.InvoiceUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=InvoiceUpdateResponse,
        )

    def list(
        self,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceListResponse:
        """Lists all invoices for the workspace"""
        return self._get(
            "/invoices",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=InvoiceListResponse,
        )

    def list_history(
        self,
        id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceListHistoryResponse:
        """
        Gets the version history of an invoice

        Args:
          id: Invoice ID

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return self._get(
            path_template("/invoices/{id}/history", id=id),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=InvoiceListHistoryResponse,
        )

    def search(
        self,
        *,
        filter: invoice_search_params.Filter,
        page_info: invoice_search_params.PageInfo,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceSearchResponse:
        """
        Searches invoices

        Args:
          filter: Filter criteria for the search

          page_info: Pagination parameters

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/invoices/search",
            body=maybe_transform(
                {
                    "filter": filter,
                    "page_info": page_info,
                },
                invoice_search_params.InvoiceSearchParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=InvoiceSearchResponse,
        )


class AsyncInvoicesResource(AsyncAPIResource):
    """Invoice management operations"""

    @cached_property
    def with_raw_response(self) -> AsyncInvoicesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/fragment-dev/fragment-py#accessing-raw-response-data-eg-headers
        """
        return AsyncInvoicesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncInvoicesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/fragment-dev/fragment-py#with_streaming_response
        """
        return AsyncInvoicesResourceWithStreamingResponse(self)

    async def create(
        self,
        *,
        invoice_id: str,
        line_items: Iterable[invoice_create_params.LineItem],
        tags: Iterable[invoice_create_params.Tag] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceCreateResponse:
        """Creates a new invoice

        Args:
          invoice_id: Unique identifier for the invoice.

        Make this the canonical ID from your system
              for the transaction.

          line_items: List of line items to create with the invoice

          tags: Optional metadata tags for this invoice

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/invoices",
            body=await async_maybe_transform(
                {
                    "invoice_id": invoice_id,
                    "line_items": line_items,
                    "tags": tags,
                },
                invoice_create_params.InvoiceCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=InvoiceCreateResponse,
        )

    async def retrieve(
        self,
        id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceRetrieveResponse:
        """
        Gets an invoice by ID with balance details

        Args:
          id: Invoice ID

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return await self._get(
            path_template("/invoices/{id}", id=id),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=InvoiceRetrieveResponse,
        )

    async def update(
        self,
        id: str,
        *,
        line_items_update: Iterable[invoice_update_params.LineItemsUpdate],
        version: float,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceUpdateResponse:
        """
        Updates an invoice with line item operations

        Args:
          id: Invoice ID

          line_items_update: List of line item operations to apply to the invoice

          version: The version of the invoice being updated. Must match the current version for the
              update to succeed.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return await self._post(
            path_template("/invoices/{id}", id=id),
            body=await async_maybe_transform(
                {
                    "line_items_update": line_items_update,
                    "version": version,
                },
                invoice_update_params.InvoiceUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=InvoiceUpdateResponse,
        )

    async def list(
        self,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceListResponse:
        """Lists all invoices for the workspace"""
        return await self._get(
            "/invoices",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=InvoiceListResponse,
        )

    async def list_history(
        self,
        id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceListHistoryResponse:
        """
        Gets the version history of an invoice

        Args:
          id: Invoice ID

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return await self._get(
            path_template("/invoices/{id}/history", id=id),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=InvoiceListHistoryResponse,
        )

    async def search(
        self,
        *,
        filter: invoice_search_params.Filter,
        page_info: invoice_search_params.PageInfo,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceSearchResponse:
        """
        Searches invoices

        Args:
          filter: Filter criteria for the search

          page_info: Pagination parameters

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/invoices/search",
            body=await async_maybe_transform(
                {
                    "filter": filter,
                    "page_info": page_info,
                },
                invoice_search_params.InvoiceSearchParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=InvoiceSearchResponse,
        )


class InvoicesResourceWithRawResponse:
    def __init__(self, invoices: InvoicesResource) -> None:
        self._invoices = invoices

        self.create = to_raw_response_wrapper(
            invoices.create,
        )
        self.retrieve = to_raw_response_wrapper(
            invoices.retrieve,
        )
        self.update = to_raw_response_wrapper(
            invoices.update,
        )
        self.list = to_raw_response_wrapper(
            invoices.list,
        )
        self.list_history = to_raw_response_wrapper(
            invoices.list_history,
        )
        self.search = to_raw_response_wrapper(
            invoices.search,
        )


class AsyncInvoicesResourceWithRawResponse:
    def __init__(self, invoices: AsyncInvoicesResource) -> None:
        self._invoices = invoices

        self.create = async_to_raw_response_wrapper(
            invoices.create,
        )
        self.retrieve = async_to_raw_response_wrapper(
            invoices.retrieve,
        )
        self.update = async_to_raw_response_wrapper(
            invoices.update,
        )
        self.list = async_to_raw_response_wrapper(
            invoices.list,
        )
        self.list_history = async_to_raw_response_wrapper(
            invoices.list_history,
        )
        self.search = async_to_raw_response_wrapper(
            invoices.search,
        )


class InvoicesResourceWithStreamingResponse:
    def __init__(self, invoices: InvoicesResource) -> None:
        self._invoices = invoices

        self.create = to_streamed_response_wrapper(
            invoices.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            invoices.retrieve,
        )
        self.update = to_streamed_response_wrapper(
            invoices.update,
        )
        self.list = to_streamed_response_wrapper(
            invoices.list,
        )
        self.list_history = to_streamed_response_wrapper(
            invoices.list_history,
        )
        self.search = to_streamed_response_wrapper(
            invoices.search,
        )


class AsyncInvoicesResourceWithStreamingResponse:
    def __init__(self, invoices: AsyncInvoicesResource) -> None:
        self._invoices = invoices

        self.create = async_to_streamed_response_wrapper(
            invoices.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            invoices.retrieve,
        )
        self.update = async_to_streamed_response_wrapper(
            invoices.update,
        )
        self.list = async_to_streamed_response_wrapper(
            invoices.list,
        )
        self.list_history = async_to_streamed_response_wrapper(
            invoices.list_history,
        )
        self.search = async_to_streamed_response_wrapper(
            invoices.search,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, TypedDict

from .._types import SequenceNotStr

__all__ = ["TransactionSearchAllocationsParams", "Filter", "FilterInvoiceID"]


class TransactionSearchAllocationsParams(TypedDict, total=False):
    filter: Required[Filter]
    """Filter criteria for searching transaction allocations."""


class FilterInvoiceID(TypedDict, total=False):
    any: Required[SequenceNotStr[str]]
    """Match allocations where invoice_id is any of these values (OR)."""


class Filter(TypedDict, total=False):
    """Filter criteria for searching transaction allocations."""

    invoice_id: Required[FilterInvoiceID]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, TypedDict

__all__ = ["ExternalAccountCreateParams"]


class ExternalAccountCreateParams(TypedDict, total=False):
    external_id: Required[str]
    """External ID for the account (user-provided, unique, mutable)"""

    name: Required[str]
    """Human-readable name for the external account (mutable)"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .._models import BaseModel

__all__ = ["Role"]


class Role(BaseModel):
    """Role object"""

    id: str
    """Unique role ID"""

    role: str
    """Name of the role"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable

import httpx

from ..types import product_create_params
from .._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from .._utils import path_template, maybe_transform, async_maybe_transform
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from .._base_client import make_request_options
from ..types.product_list_response import ProductListResponse
from ..types.product_create_response import ProductCreateResponse
from ..types.product_retrieve_response import ProductRetrieveResponse

__all__ = ["ProductsResource", "AsyncProductsResource"]


class ProductsResource(SyncAPIResource):
    """Product management operations"""

    @cached_property
    def with_raw_response(self) -> ProductsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/fragment-dev/fragment-py#accessing-raw-response-data-eg-headers
        """
        return ProductsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> ProductsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/fragment-dev/fragment-py#with_streaming_response
        """
        return ProductsResourceWithStreamingResponse(self)

    def create(
        self,
        *,
        code: str,
        description: str | Omit = omit,
        paid_by_roles: Iterable[product_create_params.PaidByRole] | Omit = omit,
        paid_to_roles: Iterable[product_create_params.PaidToRole] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> ProductCreateResponse:
        """
        Creates a new product

        Args:
          code: Product code (unique identifier)

          description: Description of the product

          paid_by_roles: Roles that can pay for this product. Reference roles by id or name. At least one
              of paid_by_roles or paid_to_roles must be provided.

          paid_to_roles: Roles that receive payment for this product. Reference roles by id or name. At
              least one of paid_by_roles or paid_to_roles must be provided.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/products",
            body=maybe_transform(
                {
                    "code": code,
                    "description": description,
                    "paid_by_roles": paid_by_roles,
                    "paid_to_roles": paid_to_roles,
                },
                product_create_params.ProductCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=ProductCreateResponse,
        )

    def retrieve(
        self,
        code: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> ProductRetrieveResponse:
        """
        Gets a product by code

        Args:
          code: Product code

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not code:
            raise ValueError(f"Expected a non-empty value for `code` but received {code!r}")
        return self._get(
            path_template("/products/{code}", code=code),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=ProductRetrieveResponse,
        )

    def list(
        self,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> ProductListResponse:
        """Lists all products for the workspace"""
        return self._get(
            "/products",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=ProductListResponse,
        )


class AsyncProductsResource(AsyncAPIResource):
    """Product management operations"""

    @cached_property
    def with_raw_response(self) -> AsyncProductsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/fragment-dev/fragment-py#accessing-raw-response-data-eg-headers
        """
        return AsyncProductsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncProductsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/fragment-dev/fragment-py#with_streaming_response
        """
        return AsyncProductsResourceWithStreamingResponse(self)

    async def create(
        self,
        *,
        code: str,
        description: str | Omit = omit,
        paid_by_roles: Iterable[product_create_params.PaidByRole] | Omit = omit,
        paid_to_roles: Iterable[product_create_params.PaidToRole] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> ProductCreateResponse:
        """
        Creates a new product

        Args:
          code: Product code (unique identifier)

          description: Description of the product

          paid_by_roles: Roles that can pay for this product. Reference roles by id or name. At least one
              of paid_by_roles or paid_to_roles must be provided.

          paid_to_roles: Roles that receive payment for this product. Reference roles by id or name. At
              least one of paid_by_roles or paid_to_roles must be provided.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/products",
            body=await async_maybe_transform(
                {
                    "code": code,
                    "description": description,
                    "paid_by_roles": paid_by_roles,
                    "paid_to_roles": paid_to_roles,
                },
                product_create_params.ProductCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=ProductCreateResponse,
        )

    async def retrieve(
        self,
        code: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> ProductRetrieveResponse:
        """
        Gets a product by code

        Args:
          code: Product code

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not code:
            raise ValueError(f"Expected a non-empty value for `code` but received {code!r}")
        return await self._get(
            path_template("/products/{code}", code=code),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=ProductRetrieveResponse,
        )

    async def list(
        self,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> ProductListResponse:
        """Lists all products for the workspace"""
        return await self._get(
            "/products",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=ProductListResponse,
        )


class ProductsResourceWithRawResponse:
    def __init__(self, products: ProductsResource) -> None:
        self._products = products

        self.create = to_raw_response_wrapper(
            products.create,
        )
        self.retrieve = to_raw_response_wrapper(
            products.retrieve,
        )
        self.list = to_raw_response_wrapper(
            products.list,
        )


class AsyncProductsResourceWithRawResponse:
    def __init__(self, products: AsyncProductsResource) -> None:
        self._products = products

        self.create = async_to_raw_response_wrapper(
            products.create,
        )
        self.retrieve = async_to_raw_response_wrapper(
            products.retrieve,
        )
        self.list = async_to_raw_response_wrapper(
            products.list,
        )


class ProductsResourceWithStreamingResponse:
    def __init__(self, products: ProductsResource) -> None:
        self._products = products

        self.create = to_streamed_response_wrapper(
            products.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            products.retrieve,
        )
        self.list = to_streamed_response_wrapper(
            products.list,
        )


class AsyncProductsResourceWithStreamingResponse:
    def __init__(self, products: AsyncProductsResource) -> None:
        self._products = products

        self.create = async_to_streamed_response_wrapper(
            products.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            products.retrieve,
        )
        self.list = async_to_streamed_response_wrapper(
            products.list,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .._models import BaseModel
from .transaction import Transaction

__all__ = ["TransactionRetrieveResponse"]


class TransactionRetrieveResponse(BaseModel):
    data: Transaction
    """Transaction object."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, TypedDict

__all__ = ["RoleCreateParams"]


class RoleCreateParams(TypedDict, total=False):
    role: Required[str]
    """Name of the role"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Iterable
from datetime import datetime
from typing_extensions import Literal

import httpx

from ..types import (
    transaction_list_params,
    transaction_create_params,
    transaction_search_params,
    transaction_create_allocations_params,
    transaction_search_allocations_params,
)
from .._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from .._utils import path_template, maybe_transform, async_maybe_transform
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from .._base_client import make_request_options
from ..types.transaction_list_response import TransactionListResponse
from ..types.transaction_create_response import TransactionCreateResponse
from ..types.transaction_search_response import TransactionSearchResponse
from ..types.transaction_retrieve_response import TransactionRetrieveResponse
from ..types.transaction_list_history_response import TransactionListHistoryResponse
from ..types.transaction_create_allocations_response import TransactionCreateAllocationsResponse
from ..types.transaction_search_allocations_response import TransactionSearchAllocationsResponse

__all__ = ["TransactionsResource", "AsyncTransactionsResource"]


class TransactionsResource(SyncAPIResource):
    """Transaction sync operations"""

    @cached_property
    def with_raw_response(self) -> TransactionsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/fragment-dev/fragment-py#accessing-raw-response-data-eg-headers
        """
        return TransactionsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> TransactionsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/fragment-dev/fragment-py#with_streaming_response
        """
        return TransactionsResourceWithStreamingResponse(self)

    def create(
        self,
        *,
        account: transaction_create_params.Account,
        allocations: Iterable[transaction_create_params.Allocation],
        amount: str,
        currency: Literal[
            "ADA",
            "BTC",
            "DAI",
            "ETH",
            "SOL",
            "USDC",
            "USDT",
            "USDG",
            "EURC",
            "CADC",
            "CADT",
            "XLM",
            "UNI",
            "BCH",
            "LTC",
            "AAVE",
            "LINK",
            "MATIC",
            "PTS",
            "AED",
            "AFN",
            "ALL",
            "AMD",
            "ANG",
            "AOA",
            "ARS",
            "AUD",
            "AWG",
            "AZN",
            "BAM",
            "BBD",
            "BDT",
            "BGN",
            "BHD",
            "BIF",
            "BMD",
            "BND",
            "BOB",
            "BRL",
            "BSD",
            "BTN",
            "BWP",
            "BYR",
            "BZD",
            "CAD",
            "CDF",
            "CHF",
            "CLP",
            "CNY",
            "COP",
            "CRC",
            "CUC",
            "CUP",
            "CVE",
            "CZK",
            "DJF",
            "DKK",
            "DOP",
            "DZD",
            "EGP",
            "ERN",
            "ETB",
            "EUR",
            "FJD",
            "FKP",
            "GBP",
            "GEL",
            "GGP",
            "GHS",
            "GIP",
            "GMD",
            "GNF",
            "GTQ",
            "GYD",
            "HKD",
            "HNL",
            "HRK",
            "HTG",
            "HUF",
            "IDR",
            "ILS",
            "IMP",
            "INR",
            "IQD",
            "IRR",
            "ISK",
            "JMD",
            "JOD",
            "JPY",
            "KES",
            "KGS",
            "KHR",
            "KMF",
            "KPW",
            "KRW",
            "KWD",
            "KYD",
            "KZT",
            "LAK",
            "LBP",
            "LKR",
            "LRD",
            "LSL",
            "LYD",
            "MAD",
            "MDL",
            "MGA",
            "MKD",
            "MMK",
            "MNT",
            "MOP",
            "MUR",
            "MVR",
            "MWK",
            "MXN",
            "MYR",
            "MZN",
            "NAD",
            "NGN",
            "NIO",
            "NOK",
            "NPR",
            "NZD",
            "OMR",
            "PAB",
            "PEN",
            "PGK",
            "PHP",
            "PKR",
            "PLN",
            "PYG",
            "QAR",
            "RON",
            "RSD",
            "RUB",
            "RWF",
            "SAR",
            "SBD",
            "SCR",
            "SDG",
            "SEK",
            "SGD",
            "SHP",
            "SLL",
            "SOS",
            "SPL",
            "SRD",
            "SVC",
            "SYP",
            "STN",
            "SZL",
            "THB",
            "TJS",
            "TMT",
            "TND",
            "TOP",
            "TRY",
            "TTD",
            "TVD",
            "TWD",
            "TZS",
            "UAH",
            "UGX",
            "USD",
            "UYU",
            "UZS",
            "VEF",
            "VND",
            "VUV",
            "WST",
            "XAF",
            "XCD",
            "XOF",
            "XPF",
            "YER",
            "ZAR",
            "ZMW",
            "LOGICAL",
            "CUSTOM",
        ],
        external_id: str,
        posted: Union[str, datetime],
        tags: Iterable[transaction_create_params.Tag] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> TransactionCreateResponse:
        """
        Syncs a transaction, optionally with allocations

        Args:
          account: Account reference. Provide id, external_id, or both.

          allocations: Allocation entries for this transaction. Empty indicates unreconciled funds.

          amount: Amount in smallest currency unit as stringified bigint (can be positive or
              negative).

          currency: Currency code (ISO 4217 or crypto)

          external_id: External transaction ID used for idempotent sync.

          posted: Posted timestamp in ISO 8601 format.

          tags: Optional metadata tags for this transaction

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/transactions",
            body=maybe_transform(
                {
                    "account": account,
                    "allocations": allocations,
                    "amount": amount,
                    "currency": currency,
                    "external_id": external_id,
                    "posted": posted,
                    "tags": tags,
                },
                transaction_create_params.TransactionCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=TransactionCreateResponse,
        )

    def retrieve(
        self,
        transaction_ref: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> TransactionRetrieveResponse:
        """
        Gets a transaction by ID or external ID

        Args:
          transaction_ref: Transaction reference. Accepts either an encoded Fragment ID (txn_xxx) or an
              external ID.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not transaction_ref:
            raise ValueError(f"Expected a non-empty value for `transaction_ref` but received {transaction_ref!r}")
        return self._get(
            path_template("/transactions/{transaction_ref}", transaction_ref=transaction_ref),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=TransactionRetrieveResponse,
        )

    def list(
        self,
        *,
        account: str | Omit = omit,
        reconciliation_status: Literal["reconciled", "unreconciled"] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> TransactionListResponse:
        """Lists all transactions for the workspace

        Args:
          account: Filter by account.

        Encoded account ID (ext_account_xxx) or external_id. If the
              account does not exist, returns an empty list.

          reconciliation_status: Filter by reconciliation state. reconciled = unallocated_amount === 0;
              unreconciled = unallocated_amount !== 0. Omit for all transactions.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get(
            "/transactions",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "account": account,
                        "reconciliation_status": reconciliation_status,
                    },
                    transaction_list_params.TransactionListParams,
                ),
            ),
            cast_to=TransactionListResponse,
        )

    def create_allocations(
        self,
        transaction_ref: str,
        *,
        allocation_updates: Iterable[transaction_create_allocations_params.AllocationUpdate],
        version: int,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> TransactionCreateAllocationsResponse:
        """
        Updates allocations on an existing transaction

        Args:
          transaction_ref: Transaction reference. Accepts either an encoded Fragment ID (txn_xxx) or an
              external ID.

          allocation_updates: Allocation operations to apply

          version: Current transaction version for optimistic concurrency control

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not transaction_ref:
            raise ValueError(f"Expected a non-empty value for `transaction_ref` but received {transaction_ref!r}")
        return self._post(
            path_template("/transactions/{transaction_ref}/allocations", transaction_ref=transaction_ref),
            body=maybe_transform(
                {
                    "allocation_updates": allocation_updates,
                    "version": version,
                },
                transaction_create_allocations_params.TransactionCreateAllocationsParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=TransactionCreateAllocationsResponse,
        )

    def list_history(
        self,
        transaction_ref: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> TransactionListHistoryResponse:
        """
        Gets the version history of a transaction

        Args:
          transaction_ref: Transaction reference. Accepts either an encoded Fragment ID (txn_xxx) or an
              external ID.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not transaction_ref:
            raise ValueError(f"Expected a non-empty value for `transaction_ref` but received {transaction_ref!r}")
        return self._get(
            path_template("/transactions/{transaction_ref}/history", transaction_ref=transaction_ref),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=TransactionListHistoryResponse,
        )

    def search(
        self,
        *,
        filter: transaction_search_params.Filter,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> TransactionSearchResponse:
        """
        Searches transactions by filter criteria

        Args:
          filter: Filter criteria for searching transactions.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/transactions/search",
            body=maybe_transform({"filter": filter}, transaction_search_params.TransactionSearchParams),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=TransactionSearchResponse,
        )

    def search_allocations(
        self,
        *,
        filter: transaction_search_allocations_params.Filter,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> TransactionSearchAllocationsResponse:
        """
        Searches transaction allocations by filter criteria

        Args:
          filter: Filter criteria for searching transaction allocations.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/transactions/allocations/search",
            body=maybe_transform(
                {"filter": filter}, transaction_search_allocations_params.TransactionSearchAllocationsParams
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=TransactionSearchAllocationsResponse,
        )


class AsyncTransactionsResource(AsyncAPIResource):
    """Transaction sync operations"""

    @cached_property
    def with_raw_response(self) -> AsyncTransactionsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/fragment-dev/fragment-py#accessing-raw-response-data-eg-headers
        """
        return AsyncTransactionsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncTransactionsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/fragment-dev/fragment-py#with_streaming_response
        """
        return AsyncTransactionsResourceWithStreamingResponse(self)

    async def create(
        self,
        *,
        account: transaction_create_params.Account,
        allocations: Iterable[transaction_create_params.Allocation],
        amount: str,
        currency: Literal[
            "ADA",
            "BTC",
            "DAI",
            "ETH",
            "SOL",
            "USDC",
            "USDT",
            "USDG",
            "EURC",
            "CADC",
            "CADT",
            "XLM",
            "UNI",
            "BCH",
            "LTC",
            "AAVE",
            "LINK",
            "MATIC",
            "PTS",
            "AED",
            "AFN",
            "ALL",
            "AMD",
            "ANG",
            "AOA",
            "ARS",
            "AUD",
            "AWG",
            "AZN",
            "BAM",
            "BBD",
            "BDT",
            "BGN",
            "BHD",
            "BIF",
            "BMD",
            "BND",
            "BOB",
            "BRL",
            "BSD",
            "BTN",
            "BWP",
            "BYR",
            "BZD",
            "CAD",
            "CDF",
            "CHF",
            "CLP",
            "CNY",
            "COP",
            "CRC",
            "CUC",
            "CUP",
            "CVE",
            "CZK",
            "DJF",
            "DKK",
            "DOP",
            "DZD",
            "EGP",
            "ERN",
            "ETB",
            "EUR",
            "FJD",
            "FKP",
            "GBP",
            "GEL",
            "GGP",
            "GHS",
            "GIP",
            "GMD",
            "GNF",
            "GTQ",
            "GYD",
            "HKD",
            "HNL",
            "HRK",
            "HTG",
            "HUF",
            "IDR",
            "ILS",
            "IMP",
            "INR",
            "IQD",
            "IRR",
            "ISK",
            "JMD",
            "JOD",
            "JPY",
            "KES",
            "KGS",
            "KHR",
            "KMF",
            "KPW",
            "KRW",
            "KWD",
            "KYD",
            "KZT",
            "LAK",
            "LBP",
            "LKR",
            "LRD",
            "LSL",
            "LYD",
            "MAD",
            "MDL",
            "MGA",
            "MKD",
            "MMK",
            "MNT",
            "MOP",
            "MUR",
            "MVR",
            "MWK",
            "MXN",
            "MYR",
            "MZN",
            "NAD",
            "NGN",
            "NIO",
            "NOK",
            "NPR",
            "NZD",
            "OMR",
            "PAB",
            "PEN",
            "PGK",
            "PHP",
            "PKR",
            "PLN",
            "PYG",
            "QAR",
            "RON",
            "RSD",
            "RUB",
            "RWF",
            "SAR",
            "SBD",
            "SCR",
            "SDG",
            "SEK",
            "SGD",
            "SHP",
            "SLL",
            "SOS",
            "SPL",
            "SRD",
            "SVC",
            "SYP",
            "STN",
            "SZL",
            "THB",
            "TJS",
            "TMT",
            "TND",
            "TOP",
            "TRY",
            "TTD",
            "TVD",
            "TWD",
            "TZS",
            "UAH",
            "UGX",
            "USD",
            "UYU",
            "UZS",
            "VEF",
            "VND",
            "VUV",
            "WST",
            "XAF",
            "XCD",
            "XOF",
            "XPF",
            "YER",
            "ZAR",
            "ZMW",
            "LOGICAL",
            "CUSTOM",
        ],
        external_id: str,
        posted: Union[str, datetime],
        tags: Iterable[transaction_create_params.Tag] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> TransactionCreateResponse:
        """
        Syncs a transaction, optionally with allocations

        Args:
          account: Account reference. Provide id, external_id, or both.

          allocations: Allocation entries for this transaction. Empty indicates unreconciled funds.

          amount: Amount in smallest currency unit as stringified bigint (can be positive or
              negative).

          currency: Currency code (ISO 4217 or crypto)

          external_id: External transaction ID used for idempotent sync.

          posted: Posted timestamp in ISO 8601 format.

          tags: Optional metadata tags for this transaction

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/transactions",
            body=await async_maybe_transform(
                {
                    "account": account,
                    "allocations": allocations,
                    "amount": amount,
                    "currency": currency,
                    "external_id": external_id,
                    "posted": posted,
                    "tags": tags,
                },
                transaction_create_params.TransactionCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=TransactionCreateResponse,
        )

    async def retrieve(
        self,
        transaction_ref: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> TransactionRetrieveResponse:
        """
        Gets a transaction by ID or external ID

        Args:
          transaction_ref: Transaction reference. Accepts either an encoded Fragment ID (txn_xxx) or an
              external ID.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not transaction_ref:
            raise ValueError(f"Expected a non-empty value for `transaction_ref` but received {transaction_ref!r}")
        return await self._get(
            path_template("/transactions/{transaction_ref}", transaction_ref=transaction_ref),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=TransactionRetrieveResponse,
        )

    async def list(
        self,
        *,
        account: str | Omit = omit,
        reconciliation_status: Literal["reconciled", "unreconciled"] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> TransactionListResponse:
        """Lists all transactions for the workspace

        Args:
          account: Filter by account.

        Encoded account ID (ext_account_xxx) or external_id. If the
              account does not exist, returns an empty list.

          reconciliation_status: Filter by reconciliation state. reconciled = unallocated_amount === 0;
              unreconciled = unallocated_amount !== 0. Omit for all transactions.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._get(
            "/transactions",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=await async_maybe_transform(
                    {
                        "account": account,
                        "reconciliation_status": reconciliation_status,
                    },
                    transaction_list_params.TransactionListParams,
                ),
            ),
            cast_to=TransactionListResponse,
        )

    async def create_allocations(
        self,
        transaction_ref: str,
        *,
        allocation_updates: Iterable[transaction_create_allocations_params.AllocationUpdate],
        version: int,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> TransactionCreateAllocationsResponse:
        """
        Updates allocations on an existing transaction

        Args:
          transaction_ref: Transaction reference. Accepts either an encoded Fragment ID (txn_xxx) or an
              external ID.

          allocation_updates: Allocation operations to apply

          version: Current transaction version for optimistic concurrency control

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not transaction_ref:
            raise ValueError(f"Expected a non-empty value for `transaction_ref` but received {transaction_ref!r}")
        return await self._post(
            path_template("/transactions/{transaction_ref}/allocations", transaction_ref=transaction_ref),
            body=await async_maybe_transform(
                {
                    "allocation_updates": allocation_updates,
                    "version": version,
                },
                transaction_create_allocations_params.TransactionCreateAllocationsParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=TransactionCreateAllocationsResponse,
        )

    async def list_history(
        self,
        transaction_ref: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> TransactionListHistoryResponse:
        """
        Gets the version history of a transaction

        Args:
          transaction_ref: Transaction reference. Accepts either an encoded Fragment ID (txn_xxx) or an
              external ID.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not transaction_ref:
            raise ValueError(f"Expected a non-empty value for `transaction_ref` but received {transaction_ref!r}")
        return await self._get(
            path_template("/transactions/{transaction_ref}/history", transaction_ref=transaction_ref),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=TransactionListHistoryResponse,
        )

    async def search(
        self,
        *,
        filter: transaction_search_params.Filter,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> TransactionSearchResponse:
        """
        Searches transactions by filter criteria

        Args:
          filter: Filter criteria for searching transactions.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/transactions/search",
            body=await async_maybe_transform({"filter": filter}, transaction_search_params.TransactionSearchParams),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=TransactionSearchResponse,
        )

    async def search_allocations(
        self,
        *,
        filter: transaction_search_allocations_params.Filter,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> TransactionSearchAllocationsResponse:
        """
        Searches transaction allocations by filter criteria

        Args:
          filter: Filter criteria for searching transaction allocations.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/transactions/allocations/search",
            body=await async_maybe_transform(
                {"filter": filter}, transaction_search_allocations_params.TransactionSearchAllocationsParams
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=TransactionSearchAllocationsResponse,
        )


class TransactionsResourceWithRawResponse:
    def __init__(self, transactions: TransactionsResource) -> None:
        self._transactions = transactions

        self.create = to_raw_response_wrapper(
            transactions.create,
        )
        self.retrieve = to_raw_response_wrapper(
            transactions.retrieve,
        )
        self.list = to_raw_response_wrapper(
            transactions.list,
        )
        self.create_allocations = to_raw_response_wrapper(
            transactions.create_allocations,
        )
        self.list_history = to_raw_response_wrapper(
            transactions.list_history,
        )
        self.search = to_raw_response_wrapper(
            transactions.search,
        )
        self.search_allocations = to_raw_response_wrapper(
            transactions.search_allocations,
        )


class AsyncTransactionsResourceWithRawResponse:
    def __init__(self, transactions: AsyncTransactionsResource) -> None:
        self._transactions = transactions

        self.create = async_to_raw_response_wrapper(
            transactions.create,
        )
        self.retrieve = async_to_raw_response_wrapper(
            transactions.retrieve,
        )
        self.list = async_to_raw_response_wrapper(
            transactions.list,
        )
        self.create_allocations = async_to_raw_response_wrapper(
            transactions.create_allocations,
        )
        self.list_history = async_to_raw_response_wrapper(
            transactions.list_history,
        )
        self.search = async_to_raw_response_wrapper(
            transactions.search,
        )
        self.search_allocations = async_to_raw_response_wrapper(
            transactions.search_allocations,
        )


class TransactionsResourceWithStreamingResponse:
    def __init__(self, transactions: TransactionsResource) -> None:
        self._transactions = transactions

        self.create = to_streamed_response_wrapper(
            transactions.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            transactions.retrieve,
        )
        self.list = to_streamed_response_wrapper(
            transactions.list,
        )
        self.create_allocations = to_streamed_response_wrapper(
            transactions.create_allocations,
        )
        self.list_history = to_streamed_response_wrapper(
            transactions.list_history,
        )
        self.search = to_streamed_response_wrapper(
            transactions.search,
        )
        self.search_allocations = to_streamed_response_wrapper(
            transactions.search_allocations,
        )


class AsyncTransactionsResourceWithStreamingResponse:
    def __init__(self, transactions: AsyncTransactionsResource) -> None:
        self._transactions = transactions

        self.create = async_to_streamed_response_wrapper(
            transactions.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            transactions.retrieve,
        )
        self.list = async_to_streamed_response_wrapper(
            transactions.list,
        )
        self.create_allocations = async_to_streamed_response_wrapper(
            transactions.create_allocations,
        )
        self.list_history = async_to_streamed_response_wrapper(
            transactions.list_history,
        )
        self.search = async_to_streamed_response_wrapper(
            transactions.search,
        )
        self.search_allocations = async_to_streamed_response_wrapper(
            transactions.search_allocations,
        )
